# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import argparse

import yaml
from configargparse import ArgumentParser, YAMLConfigFileParser
import logging
import coloredlogs
import textwrap
import sys
import os

import powerfulseal.version
from powerfulseal.k8s.metrics_server_client import MetricsServerClient
from powerfulseal.policy.demo_runner import DemoRunner
from prometheus_client import start_http_server
from powerfulseal.metriccollectors import StdoutCollector, PrometheusCollector
from powerfulseal.policy.label_runner import LabelRunner
from powerfulseal.web.server import ServerState, start_server, ServerStateLogHandler
from ..node import NodeInventory
from ..node.inventory import read_inventory_file_to_dict
from ..clouddrivers import OpenStackDriver, AWSDriver, NoCloudDriver, AzureDriver, GCPDriver
from ..execute import RemoteExecutor
from ..k8s import K8sClient, K8sInventory
from .pscmd import PSCmd
from ..policy import PolicyRunner

KUBECONFIG_DEFAULT_PATH = "~/.kube/config"

def parse_kubeconfig(args):
    """
        if explicitly set, use the --kubeconfig value
        otherwise, check if KUBECONFIG is set
        if not, check if there is `~/.kube/config` available
        else try to build in-cluster config
    """
    logger = logging.getLogger(__name__)
    kube_config = None
    expanded_home_kube_config_path = os.path.expanduser(KUBECONFIG_DEFAULT_PATH)
    if args.kubeconfig:
        kube_config = args.kubeconfig
        logger.info("Creating kubernetes client with config %s from --kubeconfig flag", kube_config)
    elif os.environ.get("KUBECONFIG"):
        kube_config = os.path.expanduser(os.environ.get("KUBECONFIG"))
        logger.info("Creating kubernetes client with config %s from KUBECONFIG env var", kube_config)
    elif os.path.exists(expanded_home_kube_config_path):
        kube_config = expanded_home_kube_config_path
        logger.info("Creating kubernetes client with config %s (path found for backwards compatibility)", kube_config)
    else:
        logger.info("Creating kubernetes client with in-cluster config")
    return kube_config

def add_kubernetes_options(parser):
    # Kubernetes
    args_kubernetes = parser.add_argument_group('Kubernetes settings')
    args_kubernetes.add_argument(
        '--kubeconfig',
        help='Location of kube-config file',
        type=os.path.expanduser
    )
    args_kubernetes.add_argument(
        '--use-pod-delete-instead-of-ssh-kill',
        help='If set, will not require SSH (will delete pods instead)',
        default=False,
        action='store_true',
    )

def add_ssh_options(parser):
    # SSH
    args_ssh = parser.add_argument_group('SSH settings')
    args_ssh.add_argument(
        '--remote-user',
        default=os.environ.get("PS_REMOTE_USER", "cloud-user"),
        help="the of the user for the ssh connections",
    )
    args_ssh.add_argument(
        '--ssh-allow-missing-host-keys',
        default=False,
        action='store_true',
        help='Allow connection to hosts not present in known_hosts',
    )
    args_ssh.add_argument(
        '--override-ssh-host',
        help=(
            'If you\'d like to execute all commands on a different host '
            '(for example for minikube) you can override it here'
        )
    )
    args_ssh.add_argument(
        '--use-private-ip',
        default=False,
        action='store_true',
        help='Use the private IP of each node (vs public IP)',
    )
    ssh_options = args_ssh.add_mutually_exclusive_group()
    ssh_options.add_argument(
        '--ssh-path-to-private-key',
        default=os.environ.get("PS_PRIVATE_KEY"),
        help='Path to ssh private key',
    )
    ssh_options.add_argument(
        '--ssh-password',
        default=os.environ.get("PS_SSH_PASSWORD"),
        help='ssh password',
    )
    args_ssh.add_argument(
        '--ssh-kill-command',
        default=os.environ.get("PS_SSH_KILL_COMMAND", "sudo docker kill -s {signal} {container_id}"),
        help='The command to execute on remote host to kill the {container_id}',
    )

def add_inventory_options(parser):
    # Inventory
    args = parser.add_argument_group('Inventory settings')
    inventory_options = args.add_mutually_exclusive_group(required=True)
    inventory_options.add_argument('-i', '--inventory-file',
        default=os.environ.get("INVENTORY_FILE"),
        help=('the inventory file (in ini format) of groups '
              'of hosts to work with')
    )
    inventory_options.add_argument('--inventory-kubernetes',
        default=os.environ.get("INVENTORY_KUBERNETES"),
        help='reads all kubernetes cluster nodes as inventory',
        action='store_true',
    )

def add_cloud_options(parser):
    # Cloud Driver
    args = parser.add_argument_group('Cloud settings')
    cloud_options = args.add_mutually_exclusive_group(required=True)
    cloud_options.add_argument('--openstack',
        default=os.environ.get("OPENSTACK_CLOUD"),
        action='store_true',
        help="use OpenStack cloud provider",
    )
    cloud_options.add_argument('--aws',
        default=os.environ.get("AWS_CLOUD"),
        action='store_true',
        help="use AWS cloud provider",
    )
    cloud_options.add_argument('--azure',
        default=os.environ.get("AZURE_CLOUD"),
        action='store_true',
        help="use Azure cloud provider",
    )
    cloud_options.add_argument('--gcp',
        default=os.environ.get("GCP_CLOUD"),
        action='store_true',
        help="use GCP cloud provider",
    )
    cloud_options.add_argument('--no-cloud',
        default=os.environ.get("NO_CLOUD"),
        action='store_true',
        help="don't use cloud provider",
    )
    # other options
    args.add_argument('--openstack-cloud-name',
        default=os.environ.get("OPENSTACK_CLOUD_NAME"),
        help="optional name of the open stack cloud from your config file to use",
    )
    args.add_argument('--azure-resource-group-name',
        default=os.environ.get("AZURE_RESORUCE_GROUP_NAME"),
        help="optional name of the Azure vm cluster resource group. Used to determine azure-node-resource-group-name.",
    )
    args.add_argument('--azure-node-resource-group-name',
        default=os.environ.get("AZURE_NODE_RESORUCE_GROUP_NAME"),
        help="name of the Azure vm cluster node resource group",
    )
    args.add_argument('--gcp-config-file',
        default=os.environ.get("GCP_CONFIG_FILE"),
        help="name of the gcloud config file (in json) to use instead of the default one",
    )

def add_namespace_options(parser):
    args = parser.add_argument_group('Kubernetes options')
    args.add_argument('--kubernetes-namespace',
        default='default',
        help='Namespace to use for label and demo mode '
             '(set to blank for all namespaces)'
    )

def add_policy_options(parser):
    # Policy
    args = parser.add_argument_group('Policy settings')
    args.add_argument('--policy-file',
        default=os.environ.get("POLICY_FILE"),
        help='the policy file to run',
        required=True
    )

def add_run_options(parser):
    # policy settings
    run_args = parser.add_argument_group(
        title='Policy settings'
    )
    run_args.add_argument('--min-seconds-between-runs',
        help='Minimum number of seconds between runs',
        default=0,
        type=int
    )
    run_args.add_argument('--max-seconds-between-runs',
        help='Maximum number of seconds between runs',
        default=300,
        type=int
    )

def add_metrics_options(parser):
    # metrics settings
    autonomous_args = parser.add_argument_group(
        title='Metrics settings'
    )
    metric_options = autonomous_args.add_mutually_exclusive_group(required=False)
    metric_options.add_argument('--stdout-collector',
        default=os.environ.get("STDOUT_COLLECTOR"),
        action='store_true',
        help="print metrics collected to stdout"
    )
    metric_options.add_argument('--prometheus-collector',
        default=os.environ.get("PROMETHEUS_COLLECTOR"),
        action='store_true',
        help="store metrics in Prometheus and expose metrics over a HTTP server"
    )

    args_prometheus = parser.add_argument_group('Prometheus settings')
    args_prometheus.add_argument(
        '--prometheus-host',
        default='127.0.0.1',
        help=(
            'Host to expose Prometheus metrics via the HTTP server when using '
            'the --prometheus-collector flag'
        ),
    )
    args_prometheus.add_argument(
        '--prometheus-port',
        default=8081,
        help=(
            'Port to expose Prometheus metrics via the HTTP server '
            'when using the --prometheus-collector flag'
        ),
        type=check_valid_port
    )

def add_common_options(parser):
    add_kubernetes_options(parser)
    add_cloud_options(parser)
    add_inventory_options(parser)
    add_ssh_options(parser)


def check_valid_port(value):
    parsed = int(value)
    min_port = 0
    max_port = 65535
    if parsed < min_port or parsed > max_port:
        raise argparse.ArgumentTypeError("%s is an invalid port number" % value)
    return parsed


def parse_args(args):
    parser = ArgumentParser(
        config_file_parser_class=YAMLConfigFileParser,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        default_config_files=['~/.config/seal', '~/.seal'],
        description=textwrap.dedent("""\
            PowerfulSeal
            The Chaos Engineering tool for Kubernetes

        """),
    )

    # General settings
    parser.add_argument(
        '-c', '--config',
        is_config_file=True,
        env_var="CONFIG",
        help='Config file path',
    )
    parser.add_argument('-v', '--verbose',
        action='count',
        help='Verbose logging.'
    )
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='%(prog)s {version}'.format(version=powerfulseal.version.__version__),
        help='Version.',
    )
    # subparsers
    subparsers = parser.add_subparsers(
        title='MODES OF OPERATION',
        description=(
            'Pick one of the following options to start the Seal in the '
            'specified mode. Learn more at '
            'https://github.com/bloomberg/powerfulseal#introduction'
        ),
        dest='mode'
    )

    ##########################################################################
    # INTERACTIVE MODE
    ##########################################################################
    parser_interactive = subparsers.add_parser('interactive',
        help=(
            'Starts an interactive CLI, which allows to manually issue '
            'commands on pods and nodes and provides a sweet autocomplete. '
            'If you\'re reading this for the first time, you should probably '
            'start here. '
            'This is a DAEMONLESS mode of operation.'
        ),
    )
    add_common_options(parser_interactive)


    ##########################################################################
    # AUTONOMOUS MODE
    ##########################################################################
    parser_autonomous = subparsers.add_parser('autonomous',
        help=(
            'Starts in autonomous mode. '
            'This is the main mode of operation. The Seal reads the policy '
            'file and executes it indefinitely. '
            'It works on nodes and pods. '
            'This is a DAEMONLESS mode of operation.'
        ),
    )
    add_common_options(parser_autonomous)
    add_policy_options(parser_autonomous)
    add_metrics_options(parser_autonomous)

    # web ui settings
    web_args = parser_autonomous.add_argument_group(
        title='Web UI settings'
    )
    web_args.add_argument(
        '--headless',
        help='Doesn\'t start the UI, just runs the policy',
        action='store_true'
    )
    web_args.add_argument(
        '--host',
        help='Specify host for the PowerfulSeal web server',
        default=os.environ.get('HOST', '127.0.0.1')
    )
    web_args.add_argument(
        '--port',
        help='Specify port for the PowerfulSeal web server',
        default=int(os.environ.get('PORT', '8080')),
        type=check_valid_port
    )

    ##########################################################################
    # LABEL MODE
    ##########################################################################
    parser_label = subparsers.add_parser('label',
        help=(
            'Starts in label mode. '
            'It reads Kubernetes pods in a specified namespace, and checks '
            ' their \'seal/*\' labels to decide which ones to kill.'
            'There is no policy needed in this mode. '
            'To learn about supported labels, read more at '
            'https://github.com/bloomberg/powerfulseal/ '
            'This is a DAEMONLESS mode of operation. '
        ),
    )
    add_common_options(parser_label)
    add_namespace_options(parser_label)
    add_run_options(parser_label)
    add_metrics_options(parser_label)

    ##########################################################################
    # DEMO MODE
    ##########################################################################
    parser_demo = subparsers.add_parser('demo',
        help=(
            'Starts in demo mode. '
            'It reads Kubernetes pods in specified namespaces, and reads '
            'metrics-server metrics to guess what\'s worth killing. '
            'This is a DAEMONLESS mode of operation. '
        ),
    )
    add_common_options(parser_demo)
    add_namespace_options(parser_demo)
    add_run_options(parser_demo)
    add_metrics_options(parser_demo)

    demo_options = parser_demo.add_argument_group(
        title='metrics-server settings'
    )
    demo_options.add_argument('--metrics-server-path',
        help='Base path of metrics-server without trailing slash',
        required=True
    )
    demo_options.add_argument('--aggressiveness',
        help='Aggressiveness of demo mode (default: 3)',
        default=3,
        type=int
    )

    ##########################################################################
    # VALIDATE POLICY MODE
    ##########################################################################
    parser_validate_policy = subparsers.add_parser('validate',
        help=(
            'Validates any file against the policy schema, returns.'
            'You can use this to check that your policy is correct, '
            'before using it in autonomous mode.'
        )
    )
    add_policy_options(parser_validate_policy)

    return parser.parse_args(args=args)


def main(argv):
    """
        The main function to invoke the powerfulseal cli
    """
    args = parse_args(args=argv)

    if args.mode is None:
        return parse_args(['--help'])

    ##########################################################################
    # VALIDATE POLICY MODE
    ##########################################################################
    if args.mode == 'validate':
        policy = PolicyRunner.load_file(args.policy_file)
        if PolicyRunner.is_policy_valid(policy):
            return print('OK')
        print("Policy not valid. See log output above.")
        return os.exit(1)

    ##########################################################################
    # LOGGING
    ##########################################################################
    # Ensure the logger config propagates from the root module of this package
    logger = logging.getLogger(__name__.split('.')[0])

    # The default level should be set to logging.DEBUG to ensure that the stdout
    # stream handler can filter to the user-specified verbosity level while the
    # server logging handler can receive all logs
    logger.setLevel(logging.DEBUG)

    # Configure logging for stdout
    if not args.verbose:
        log_level = logging.ERROR
    elif args.verbose == 1:
        log_level = logging.WARNING
    elif args.verbose == 2:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    stdout_handler = logging.StreamHandler()
    stdout_handler.setLevel(log_level)
    logger.addHandler(stdout_handler)
    coloredlogs.install(logger=logger)

    my_verb = args.verbose
    logger.info("modules %s : verbosity %s : log level %s : handler level %s ", __name__, my_verb, logging.getLevelName(logger.getEffectiveLevel()), logging.getLevelName(log_level) )

    ##########################################################################
    # KUBERNETES
    ##########################################################################
    kube_config = parse_kubeconfig(args)
    k8s_client = K8sClient(kube_config=kube_config)
    k8s_inventory = K8sInventory(
        k8s_client=k8s_client,
        delete_pods=args.use_pod_delete_instead_of_ssh_kill
    )

    ##########################################################################
    # CLOUD DRIVER
    ##########################################################################
    if args.openstack:
        logger.info("Building OpenStack driver")
        driver = OpenStackDriver(
            cloud=args.openstack_cloud_name,
        )
    elif args.aws:
        logger.info("Building AWS driver")
        driver = AWSDriver()
    elif args.azure:
        logger.info("Building Azure driver")
        driver = AzureDriver(
            cluster_rg_name=args.azure_resource_group_name,
            cluster_node_rg_name=args.azure_node_resource_group_name,
        )
    elif args.gcp:
        logger.info("Building GCP driver")
        driver = GCPDriver(config=args.gcp_config_file)
    else:
        logger.info("No driver - some functionality disabled")
        driver = NoCloudDriver()

    ##########################################################################
    # INVENTORY
    ##########################################################################
    if args.inventory_file:
        logger.info("Reading inventory from %s", args.inventory_file)
        groups_to_restrict_to = read_inventory_file_to_dict(
            args.inventory_file
        )
    else:
        logger.info("Attempting to read the inventory from kubernetes")
        groups_to_restrict_to = k8s_client.get_nodes_groups()

    logger.debug("Restricting inventory to %s" % groups_to_restrict_to)

    inventory = NodeInventory(
        driver=driver,
        restrict_to_groups=groups_to_restrict_to,
    )
    inventory.sync()

    ##########################################################################
    # SSH EXECUTOR
    ##########################################################################
    if args.use_private_ip:
        logger.info("Using each node's private IP address")
    executor = RemoteExecutor(
        user=args.remote_user,
        ssh_allow_missing_host_keys=args.ssh_allow_missing_host_keys,
        ssh_path_to_private_key=args.ssh_path_to_private_key,
        override_host=args.override_ssh_host,
        ssh_password=args.ssh_password,
        use_private_ip=args.use_private_ip,
        ssh_kill_command=args.ssh_kill_command,
    )

    ##########################################################################
    # INTERACTIVE MODE
    ##########################################################################
    if args.mode == 'interactive':
        # create a command parser
        cmd = PSCmd(
            inventory=inventory,
            driver=driver,
            executor=executor,
            k8s_inventory=k8s_inventory,
        )
        logger.info("STARTING INTERACTIVE MODE")
        while True:
            try:
                cmd.cmdloop()
            except GeneratorExit:
                print("Exiting")
                sys.exit(0)
            except KeyboardInterrupt:
                print()
                print("Ctrl-c again to quit")
            try:
                six.moves.input()
            except KeyboardInterrupt:
                sys.exit(0)
        return

    ##########################################################################
    # METRICS
    ##########################################################################
    metric_collector = StdoutCollector()
    if args.prometheus_collector:
        logger.info("Starting prometheus metrics server on %s", args.prometheus_port)
        start_http_server(args.prometheus_port, args.prometheus_host)
        metric_collector = PrometheusCollector()
    else:
        logger.info("Not starting prometheus collector")

    ##########################################################################
    # AUTONOMOUS MODE
    ##########################################################################
    if args.mode == 'autonomous':

        # read and validate the policy
        policy = PolicyRunner.load_file(args.policy_file)
        if not PolicyRunner.is_policy_valid(policy):
            logger.info("Policy not valid. See log output above.")
            return os.exit(1)

        # run the metrics server if requested
        if not args.headless:
            # Create an instance of the singleton server state, ensuring all logs
            # for retrieval from the web interface
            state = ServerState(
                policy,
                inventory,
                k8s_inventory,
                driver,
                executor,
                args.host,
                args.port,
                args.policy_file,
                metric_collector=metric_collector,
            )
            server_log_handler = ServerStateLogHandler()
            server_log_handler.setLevel(log_level)
            logger.addHandler(server_log_handler)
            state.start_policy_runner()
            # start the server
            logger.info("Starting the UI server")
            start_server(args.host, args.port)
        else:
            logger.info("NOT starting the UI server")

            logger.info("STARTING AUTONOMOUS MODE")
            PolicyRunner.run(
                policy,
                inventory,
                k8s_inventory,
                driver,
                executor,
                metric_collector=metric_collector
            )

    ##########################################################################
    # LABEL MODE
    ##########################################################################
    elif args.mode == 'label':
        label_runner = LabelRunner(
            inventory,
            k8s_inventory,
            driver,
            executor,
            min_seconds_between_runs=args.min_seconds_between_runs,
            max_seconds_between_runs=args.max_seconds_between_runs,
            namespace=args.kubernetes_namespace,
            metric_collector=metric_collector,
        )
        logger.info("STARTING LABEL MODE")
        label_runner.run()

    ##########################################################################
    # DEMO MODE
    ##########################################################################
    elif args.mode == 'demo':
        aggressiveness = int(args.aggressiveness)
        if not 1 <= aggressiveness <= 5:
            print("Aggressiveness must be between 1 and 5 inclusive")
            os.exit(1)

        metrics_server_client = MetricsServerClient(args.metrics_server_path)
        demo_runner = DemoRunner(
            inventory,
            k8s_inventory,
            driver,
            executor,
            metrics_server_client,
            aggressiveness=aggressiveness,
            min_seconds_between_runs=args.min_seconds_between_runs,
            max_seconds_between_runs=args.max_seconds_between_runs,
            namespace=args.kubernetes_namespace,
            metric_collector=metric_collector,
        )
        logger.info("STARTING DEMO MODE")
        demo_runner.run()


def start():
    main(sys.argv[1:])


if __name__ == '__main__':
    start()




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
import re
from datetime import datetime
import calendar
import random
import logging
import abc

from powerfulseal.metriccollectors.stdout_collector import StdoutCollector


class Scenario():
    """ Basic class to represent a single testing scenario.

        Scenarios consist of 3 lists of things:
            - matches - to create the intial set of items
            - filters - to filter out the set
            - actions - to execute on all of the remaining items

        The scenarios are described using a yaml schema, that conforms
        to powerfulseal/policy/ps-schema.json JSON schema.

        This is a base class, containing some shared filters, shouldn't be
        used by itself. It's extended for both node and pod scenarios.
    """

    def __init__(self, name, schema, logger=None, metric_collector=None):
        self.name = name
        self.schema = schema
        self.logger = logger or logging.getLogger(__name__ + "." + name)
        self.metric_collector = metric_collector or StdoutCollector()
        self.property_rewrite = {
            "group": "groups",
        }

    def execute(self):
        """ Main entry point to starting a scenario.

            It calls .match() to compute the intial set of items,
            then goes through all the filters in sequence,
            and finally executes all the actions on all remaining items.
        """
        initial_set = self.match()
        self.logger.debug("Initial set: %r", initial_set)
        self.logger.info("Initial set length: %d", len(initial_set))
        filtered_set = self.filter(initial_set)
        self.logger.debug("Filtered set: %r", filtered_set)
        self.logger.info("Filtered set length: %d", len(filtered_set))
        self.act(filtered_set)
        self.logger.debug("Done")

    @abc.abstractmethod
    def match(self):
        """ Reads the policy and returns the initial set of items.
        """
        pass # pragma: no cover

    def match_property(self, candidate, criterion):
        """ Helper method to match a property following some criterion.
            Turns the value into a regular expression.
        """
        if not criterion:
            return False
        attr = criterion.get("name")
        attr = self.property_rewrite.get(attr, attr)
        value = getattr(candidate, attr)
        expr = re.compile(criterion.get("value"))
        if type(value) is list:
            return any([
                expr.match(str(v))
                for v in value
            ])
        else:
            value = str(value)
        return expr.match(value)

    def filter(self, items):
        """ Applies various filters based on the given policy.
        """
        filters = self.schema.get("filters", [])
        mapping = {
            "property": self.filter_property,
            "dayTime": self.filter_day_time,
            "randomSample": self.filter_random_sample,
            "probability": self.filter_probability,
        }
        return self.filter_mapping(items, filters, mapping)

    def filter_property(self, candidates, criterion):
        """ Filters out things which don't match their property filters.
        """
        return [
            candidate for candidate in candidates
            if self.match_property(candidate, criterion)
        ]

    def filter_day_time(self, candidates, criterion, now=None):
        """ Passed unchanged list of candidates, if the execution time
            satisfies the policy requirements.
        """
        now = now or datetime.now()
        self.logger.info("Now is %r", now)

        # check the day is permitted
        day_name = calendar.day_name[now.weekday()].lower()
        permitted_days = criterion.get("onlyDays", [])
        if permitted_days and day_name not in permitted_days:
            self.logger.info("Not allowed on %s", day_name)
            return []

        # check the time is not too early
        start = criterion.get("startTime", {})
        start_date = now.replace(
            hour=start.get("hour", 10),
            minute=start.get("minute", 0),
            second=start.get("second", 0),
        )
        if now < start_date:
            self.logger.info("Too early")
            return []

        # check the time is not too late
        end = criterion.get("endTime", {})
        end_date = now.replace(
            hour=end.get("hour", 15),
            minute=end.get("minute", 59),
            second=end.get("second", 59),
        )
        if now > end_date:
            self.logger.info("Too late")
            return []

        return candidates

    def filter_random_sample(self, candidates, criterion):
        """ Returns a random sample from the initial list.
            It supports policy `size` and `ratio` features.
        """
        if not criterion:
            return []
        size = criterion.get("size")
        if size is None:
            ratio = criterion.get("ratio", 1)
            size = int(len(candidates)*ratio)
        if size == 0:
            self.logger.info("RandomSample size 0")
            return []
        return random.sample(candidates, size)

    def filter_probability(self, candidates, criterion):
        """ Returns the initial set unchanged with given probability.
            Returns empty list otherwise.
        """
        proba = float(criterion.get("probabilityPassAll", 0.5))
        if random.random() > proba:
            self.metric_collector.add_probability_filter_passed_no_nodes_filter()
            return []
        return candidates

    def filter_mapping(self, items, filters, mapping):
        """ Executes filters mapped to methods, based on policy keywords.
        """
        for criterion in filters:
            filter_method = None
            filter_params = None
            for filter_type in mapping.keys():
                if filter_type in criterion:
                    filter_method = mapping.get(filter_type)
                    filter_params = criterion.get(filter_type)
                    len_before = len(items)
                    items = filter_method(items, filter_params)
                    len_after = len(items)
                    self.logger.debug("Filter %s: %d -> %d items", filter_type, len_before, len_after)
                    break
            if not items:
                self.logger.info("Empty set after %r", criterion)
                break

        if not items:
            self.metric_collector.add_filtered_to_empty_set_metric()

        return items

    @abc.abstractmethod
    def act(self, items):
        """ Execute policy's actions on the items,
        """
        pass # pragma: no cover

    def action_wait(self, item, params):
        """ Waits x seconds, according to the policy.
        """
        sleep_time = params.get("seconds", 0)
        self.logger.info("Action sleep for %s seconds", sleep_time)
        time.sleep(sleep_time)

    def act_mapping(self, items, actions, mapping):
        """ Executes all the actions on the list of pods.
        """
        for action in actions:
            for key, method in mapping.items():
                if key in action:
                    params = action.get(key)
                    for item in items:
                        method(item, params)
                        # special case - if we're waiting, only do that on first item
                        if key == "wait":
                            break





# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random

import pytest
from mock import MagicMock

from powerfulseal.execute import RemoteExecutor
from powerfulseal.node import NodeInventory, Node
from powerfulseal.policy.demo_runner import DemoRunner


def test_kill_pod():
    demo_runner = DemoRunner(NodeInventory(None), None, None, RemoteExecutor(), None)

    # Patch action of getting nodes to execute kill command on
    test_node = Node(1)
    get_node_by_ip_mock = MagicMock(return_value=test_node)
    demo_runner.inventory.get_node_by_ip = get_node_by_ip_mock

    # Patch action of choosing container
    execute_mock = MagicMock()
    demo_runner.executor.execute = execute_mock

    mock_pod = MagicMock()
    mock_pod.container_ids = ["docker://container1"]
    demo_runner.kill_pod(mock_pod)

    mock_pod = MagicMock()
    mock_pod.container_ids = ["docker://container1"]
    demo_runner.kill_pod(mock_pod)

    execute_mock.assert_called_with("sudo docker kill -s SIGTERM container1", nodes=[test_node])


def test_sort_pods():
    pod_1 = MagicMock()
    pod_1.metrics = {
        'cpu': 10,
        'memory': 0
    }

    pod_2 = MagicMock()
    pod_2.metrics = {
        'cpu': 5,
        'memory': 3
    }

    pod_3 = MagicMock()
    pod_3.metrics = {
        'cpu': 5,
        'memory': 2
    }

    demo_runner = DemoRunner(NodeInventory(None), None, None, RemoteExecutor(), None)
    assert demo_runner.sort_pods([pod_2, pod_3, pod_1]) == [pod_1, pod_2, pod_3]


def test_filter_top():
    first_quintile_runner = DemoRunner(None, None, None, None, None, aggressiveness=1)
    assert len(first_quintile_runner.filter_top([MagicMock() for _ in range(5)])) == 1
    assert len(first_quintile_runner.filter_top([MagicMock() for _ in range(10)])) == 2
    assert len(first_quintile_runner.filter_top([MagicMock() for _ in range(3)])) == 0
    
    third_quintile_runner = DemoRunner(None, None, None, None, None, aggressiveness=3)
    assert len(third_quintile_runner.filter_top([MagicMock() for _ in range(5)])) == 3
    assert len(third_quintile_runner.filter_top([MagicMock() for _ in range(10)])) == 6
    assert len(third_quintile_runner.filter_top([MagicMock() for _ in range(3)])) == 1

    all_runner = DemoRunner(NodeInventory(None), None, None, None, None, aggressiveness=5)
    assert len(all_runner.filter_top([MagicMock() for _ in range(5)])) == 5
    assert len(all_runner.filter_top([MagicMock() for _ in range(10)])) == 10
    assert len(all_runner.filter_top([MagicMock() for _ in range(3)])) == 3


@pytest.mark.parametrize('case', [
    (1, 0.1),
    (2, 0.275),
    (3, 0.45),
    (4, 0.625),
    (5, 0.8)
])
def test_filter_probability(case):
    random.seed(12)
    sample = 1000

    runner = DemoRunner(None, None, None, None, None, aggressiveness=case[0])
    filtered_items = runner.filter_probability([MagicMock() for _ in range(sample)])
    assert pytest.approx(case[1] * sample, 1) == len(filtered_items)




import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

long_description = """
**PowerfulSeal** adds chaos to your Kubernetes clusters, so that you can detect problems in your systems as early as possible. It kills targeted pods and takes VMs up and down.

Please see the documentation at https://github.com/bloomberg/powerfulseal
"""

setup(
    name='powerfulseal',
    version='2.6.0',
    author='Mikolaj Pawlikowski',
    author_email='mikolaj@pawlikowski.pl',
    url='https://github.com/bloomberg/powerfulseal',
    packages=find_packages(),
    license=read('LICENSE'),
    description='PowerfulSeal - a powerful testing tool for Kubernetes clusters',
    long_description=long_description,
    #long_description_content_type="text/markdown",
    install_requires=[
        'ConfigArgParse>=0.11.0,<1',
        'Flask>=1.0.0,<2',
        'termcolor>=1.1.0,<2',
        'openstacksdk>=0.10.0,<1',
        'spur>=0.3.20,<1',
        'kubernetes>=8.0.1<9',
        'PyYAML>=5.1.2,<6',
        'jsonschema>=3.0.2,<4',
        'boto3>=1.5.15,<2.0.0',
        'azure-common>=1.1.23,<2.0.0',
        'azure-mgmt-resource>=2.2.0,<3.0.0',
        'azure-mgmt-network>=2.7.0,<3.0.0',
        'azure-mgmt-compute>=4.6.2,<5.0.0',
        'future>=0.16.0,<1',
        'requests>=2.21.0,<3',
        'prometheus_client>=0.3.0,<0.4.0',
        'flask-cors>=3.0.6,<4',
        'flask-swagger-ui>=3.18.0<4',
        'coloredlogs>=10.0.0,<11.0.0',
        'six>=1.12.0,<2',
        'paramiko>=2.5.0,<3',
        'google-api-python-client>=1.7.8',
        'google-auth>=1.6.2',
        'google-auth-httplib2>=0.0.3',
        'oauth2client>=4.1.3'
    ],
    extras_require={
        'testing': [
            'pytest>=3.0,<4',
            'pytest-cov>=2.5,<3',
            'mock>=2,<3',
        ]
    },
    entry_points={
        'console_scripts': [
            'seal = powerfulseal.cli.__main__:start',
            'powerfulseal = powerfulseal.cli.__main__:start',
        ]
    },
    classifiers=[
	    'Development Status :: 4 - Beta',
	    'Intended Audience :: Developers',
	    'Topic :: Software Development :: Test Tools',
	    'Programming Language :: Python :: 3',
    ],
    python_requires='>=3',
    include_package_data=True,
)




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from . import AbstractDriver
from ..node import Node, NodeState

MESSAGE_IM_NO_CLOUD_DRIVER = (
"Trying to %s things while using a no-cloud driver. "
"If you don't expect to be seeing this, you might want"
" to rethink some of your choices"
)


class NoCloudDriver(AbstractDriver):
    """
        Concrete implementation of a noop driver
    """

    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)

    def sync(self):
        """ Noop
        """
        self.logger.error(
            MESSAGE_IM_NO_CLOUD_DRIVER, "sync"
        )

    def get_by_ip(self, ip):
        """ Creates a Node instance for given IP.
        """
        return Node(
            id="fake-{ip}".format(ip=ip),
            ip=ip,
            extIp=ip,
            az="nope",
            name="local-{ip}".format(ip=ip),
            state=NodeState.UNKNOWN
        )

    def stop(self, node):
        """ Noop
        """
        self.logger.error(
            MESSAGE_IM_NO_CLOUD_DRIVER, "stop"
        )

    def start(self, node):
        """ Noop
        """
        self.logger.error(
            MESSAGE_IM_NO_CLOUD_DRIVER, "start"
        )

    def delete(self, node):
        """ Noop
        """
        self.logger.error(
            MESSAGE_IM_NO_CLOUD_DRIVER, "delete"
        )




import pytest
from mock import patch, MagicMock
from powerfulseal.clouddrivers import gcp_driver
from powerfulseal.node import Node

PRIVATE_IPS = ['198.168.1.1', '198.168.2.1']
PUBLIC_IPS = ['31.31.31.31', '41.41.41.41']
INVALID_IP = '198.168.3.1'

GCLOUD_CONFIG = '''
{
  "compute": {
    "region": "us-central1"
  },
  "core": {
    "project": "marine-foundry-184612"
  }
}
'''


class Computeinstance():
    def __init__(
            self,
            id,
            private_ip_address,
            public_ip_address,
            zone,
            state,
            name):
        self.private_ip_address = private_ip_address
        self.id = id
        self.public_ip_address = public_ip_address
        self.placement = zone
        self.state = state
        self.name = name

    def show_as_dict(self):
        return {'name': self.name,
                'id': self.id,
                'zone': 'url/' + self.placement,
                'status': self.state,
                'networkInterfaces': [{'networkIP': self.private_ip_address,
                                       'accessConfigs': [{'natIP': self.public_ip_address}]}]}


@pytest.fixture
def compute_instances():
    return [
        Computeinstance(
            id="6285420104283695093",
            private_ip_address="198.168.1.1",
            zone="us-central1",
            public_ip_address="31.31.31.31",
            state="RUNNING",
            name="gke--cluster-1-pool-1-902c09f2-l5hf").show_as_dict(),
        Computeinstance(
            id="5611252125784513532",
            private_ip_address="198.168.2.1",
            zone="us-central2",
            public_ip_address="41.41.41.41",
            state="RUNNING",
            name="gke--cluster-1-pool-1-906c09e4-l9gh").show_as_dict()]


@patch('powerfulseal.clouddrivers.gcp_driver.create_connection_from_config')
def test_get_by_ip_private_ip_nodes(
        create_connection_from_config,
        compute_instances):
    driver = gcp_driver.GCPDriver(config=GCLOUD_CONFIG)
    driver.remote_servers = compute_instances
    node_index = 0
    nodes = driver.get_by_ip(PRIVATE_IPS[node_index])
    assert compute_instances[node_index]['id'] is nodes.id
    assert compute_instances[node_index]['zone'] == 'url/' + nodes.az
    assert compute_instances[node_index]['networkInterfaces'][0]['networkIP'] == nodes.ip


@patch('powerfulseal.clouddrivers.gcp_driver.create_connection_from_config')
def test_get_by_ip_public_ip_nodes(
        create_connection_from_config,
        compute_instances):
    driver = gcp_driver.GCPDriver(config=GCLOUD_CONFIG)
    driver.remote_servers = compute_instances
    node_index = 1
    nodes = driver.get_by_ip(PUBLIC_IPS[node_index])
    assert compute_instances[node_index]['id'] is nodes.id
    assert compute_instances[node_index]['zone'] == 'url/' + nodes.az
    assert compute_instances[node_index]['networkInterfaces'][0]['accessConfigs'][0]['natIP'] == nodes.extIp


@patch('powerfulseal.clouddrivers.gcp_driver.create_connection_from_config')
def test_get_by_ip_no_nodes(create_connection_from_config, compute_instances):
    driver = gcp_driver.GCPDriver(config=GCLOUD_CONFIG)
    driver.remote_servers = compute_instances
    nodes = driver.get_by_ip(INVALID_IP)
    assert nodes is None



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from prometheus_client import Counter

from powerfulseal.metriccollectors import AbstractCollector
from powerfulseal.metriccollectors.collector import NODE_SOURCE, POD_SOURCE

STATUS_SUCCESS = 'success'
STATUS_FAILURE = 'failure'

# Define Prometheus metrics to be stored in the default registry
POD_KILLS_METRIC_NAME = 'seal_pod_kills_total'
POD_KILLS = Counter(POD_KILLS_METRIC_NAME,
                    'Number of pods killed (including failures)',
                    ['status', 'namespace', 'name'])

NODE_STOPS_METRIC_NAME = 'seal_nodes_stopped_total'
NODE_STOPS = Counter(NODE_STOPS_METRIC_NAME,
                     'Number of nodes stopped (including failures)',
                     ['status', 'id', 'name'])

EXECUTE_FAILED_METRIC_NAME = 'seal_execute_failed_total'
EXECUTE_FAILURES = Counter(EXECUTE_FAILED_METRIC_NAME,
                           'Increasing counter for command execution failures',
                           ['id', 'name'])

FILTERED_TO_EMPTY_SET_METRIC_NAME = 'seal_empty_filter_total'
FILTERED_TO_EMPTY_SET = Counter(FILTERED_TO_EMPTY_SET_METRIC_NAME,
                                'Increasing counter for cases where filtering '
                                'returns an empty result')

PROBABILITY_FILTER_NOT_PASSED_METRIC_NAME = 'seal_probability_filter_not_passed_total'
PROBABILITY_FILTER_NOT_PASSED = Counter(PROBABILITY_FILTER_NOT_PASSED_METRIC_NAME,
                                        'Increasing counter for cases where the'
                                        ' probability filter does not pass any '
                                        'nodes')

MATCHED_TO_EMPTY_SET_METRIC_NAME = 'seal_empty_match_total'
MATCHED_TO_EMPTY_SET = Counter(MATCHED_TO_EMPTY_SET_METRIC_NAME,
                               'Increasing counter for cases where matching '
                               'returns an empty result',
                               ['source'])


class PrometheusCollector(AbstractCollector):
    def __init__(self):
        # Export 0 for time series metrics which have labels which can have default
        # values filled to avoid missing metrics. The Prometheus Python library
        # already exports 0 for metrics which do not have any labels.
        MATCHED_TO_EMPTY_SET.labels(NODE_SOURCE).inc(0)
        MATCHED_TO_EMPTY_SET.labels(POD_SOURCE).inc(0)

    def add_pod_killed_metric(self, pod):
        POD_KILLS.labels(STATUS_SUCCESS, pod.namespace, pod.name).inc()

    def add_pod_kill_failed_metric(self, pod):
        POD_KILLS.labels(STATUS_FAILURE, pod.namespace, pod.name).inc()

    def add_node_stopped_metric(self, node):
        NODE_STOPS.labels(STATUS_SUCCESS, node.id, node.name).inc()

    def add_node_stop_failed_metric(self, node):
        NODE_STOPS.labels(STATUS_FAILURE, node.id, node.name).inc()

    def add_execute_failed_metric(self, node):
        EXECUTE_FAILURES.labels(node.id, node.name).inc()

    def add_filtered_to_empty_set_metric(self):
        FILTERED_TO_EMPTY_SET.inc()

    def add_probability_filter_passed_no_nodes_filter(self):
        PROBABILITY_FILTER_NOT_PASSED.inc()

    def add_matched_to_empty_set_metric(self, source):
        MATCHED_TO_EMPTY_SET.labels(source).inc()
